import torch
import numpy as np

from . import dndarray
from . import factories
from . import operations
from . import stride_tricks
from . import types
from .communication import MPI

__all__ = [
    'expand_dims',
    'squeeze']


def expand_dims(a, axis):
    """
    Expand the shape of an array.

    Insert a new axis that will appear at the axis position in the expanded array shape.

    Parameters
    ----------
    a : ht.DNDarray
        Input array to be expanded.
    axis : int
        Position in the expanded axes where the new axis is placed.

    Returns
    -------
    res : ht.DNDarray
        Output array. The number of dimensions is one greater than that of the input array.

    Raises
    ------
    ValueError
        If the axis is not in range of the axes.

    Examples
    --------
    >>> x = ht.array([1,2])
    >>> x.shape
    (2,)

    >>> y = ht.expand_dims(x, axis=0)
    >>> y
    array([[1, 2]])
    >>> y.shape
    (1, 2)

    y = ht.expand_dims(x, axis=1)
    >>> y
    array([[1],
           [2]])
    >>> y.shape
    (2, 1)
    """
    # ensure type consistency
    if not isinstance(a, dndarray.DNDarray):
        raise TypeError('expected ht.DNDarray, but was {}'.format(type(a)))

    # sanitize axis, introduce arbitrary dummy dimension to model expansion
    axis = stride_tricks.sanitize_axis(a.shape + (1,), axis)

    return dndarray.DNDarray(
        a._DNDarray__array.unsqueeze(dim=axis), a.shape[:axis] + (1,) + a.shape[axis:],
        a.dtype,
        a.split if a.split is None or a.split < axis else a.split + 1,
        a.device,
        a.comm
    )


def squeeze(x, axis=None):
    """
    Remove single-dimensional entries from the shape of a tensor.

    Parameters:
    -----------
    x : ht.DNDarray
        Input data.

    axis : None or int or tuple of ints, optional
           Selects a subset of the single-dimensional entries in the shape.
           If axis is None, all single-dimensional entries will be removed from the shape.
           If an axis is selected with shape entry greater than one, a ValueError is raised.


    Returns:
    --------
    squeezed : ht.DNDarray
               The input tensor, but with all or a subset of the dimensions of length 1 removed.


    Examples:
    >>> import heat as ht
    >>> import torch
    >>> torch.manual_seed(1)
    <torch._C.Generator object at 0x115704ad0>
    >>> a = ht.random.randn(1,3,1,5)
    >>> a
    tensor([[[[ 0.2673, -0.4212, -0.5107, -1.5727, -0.1232]],

            [[ 3.5870, -1.8313,  1.5987, -1.2770,  0.3255]],

            [[-0.4791,  1.3790,  2.5286,  0.4107, -0.9880]]]])
    >>> a.shape
    (1, 3, 1, 5)
    >>> ht.squeeze(a).shape
    (3, 5)
    >>> ht.squeeze(a)
    tensor([[ 0.2673, -0.4212, -0.5107, -1.5727, -0.1232],
            [ 3.5870, -1.8313,  1.5987, -1.2770,  0.3255],
            [-0.4791,  1.3790,  2.5286,  0.4107, -0.9880]])
    >>> ht.squeeze(a,axis=0).shape
    (3, 1, 5)
    >>> ht.squeeze(a,axis=-2).shape
    (1, 3, 5)
    >>> ht.squeeze(a,axis=1).shape
    Traceback (most recent call last):
    ...
    ValueError: Dimension along axis 1 is not 1 for shape (1, 3, 1, 5)
    """

    # Sanitize input
    if not isinstance(x, dndarray.DNDarray):
        raise TypeError('expected x to be a ht.DNDarray, but was {}'.format(type(x)))
    # Sanitize axis
    axis = stride_tricks.sanitize_axis(x.shape, axis)
    if axis is not None:
        if isinstance(axis, int):
            dim_is_one = (x.shape[axis] == 1)
        if isinstance(axis, tuple):
            dim_is_one = bool(factories.array(list(x.shape[dim] == 1 for dim in axis)).all()._DNDarray__array)
        if not dim_is_one:
            raise ValueError('Dimension along axis {} is not 1 for shape {}'.format(axis, x.shape))

    # Local squeeze
    if axis is None:
        axis = tuple(i for i, dim in enumerate(x.shape) if dim == 1)
    if isinstance(axis, int):
        axis = (axis,)
    out_lshape = tuple(x.lshape[dim] for dim in range(len(x.lshape)) if not dim in axis)
    x_lsqueezed = x._DNDarray__array.reshape(out_lshape)

    # Calculate split axis according to squeezed shape
    if x.split is not None:
        split = x.split - len(list(dim for dim in axis if dim < x.split))
    else:
        split = x.split

    # Distributed squeeze
    if x.split is not None:
        if x.comm.is_distributed():
            if x.split in axis:
                raise ValueError('Cannot split AND squeeze along same axis. Split is {}, axis is {} for shape {}'.format(
                    x.split, axis, x.shape))
            out_gshape = tuple(x.gshape[dim] for dim in range(len(x.gshape)) if not dim in axis)
            x_gsqueezed = factories.empty(out_gshape, dtype=x.dtype)
            loffset = factories.zeros(1, dtype=types.int64)
            loffset.__setitem__(0, x.comm.chunk(x.gshape, x.split)[0])
            displs = factories.zeros(x.comm.size, dtype=types.int64)
            x.comm.Allgather(loffset, displs)

            # TODO: address uneven distribution of dimensions (Allgatherv). Issue #273, #233
            x.comm.Allgather(x_lsqueezed, x_gsqueezed)  # works with evenly distributed dimensions only
            return dndarray.DNDarray(
                x_gsqueezed,
                out_gshape,
                x_lsqueezed.dtype,
                split=split,
                device=x.device,
                comm=x.comm)

    return dndarray.DNDarray(
        x_lsqueezed,
        out_lshape,
        x.dtype,
        split=split,
        device=x.device,
        comm=x.comm)



import sys

import heat as ht


class KMeans:
    def __init__(self, n_clusters, max_iter=1000, tol=1e-4, random_state=42):
        # TODO: document me
        # TODO: extend the parameter list
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state

    @staticmethod
    def initialize_centroids(k, dimensions, seed, device):
        # TODO: document me
        # TODO: extend me with further initialization methods
        # zero-centered uniform random distribution in [-1, 1]
        ht.random.set_gseed(seed)
        return ht.random.uniform(low=-1.0, high=1.0, size=(1, dimensions, k), device=device)

    def fit(self, data):
        # TODO: document me
        data = data.expand_dims(axis=2)

        # initialize the centroids randomly
        centroids = self.initialize_centroids(
            self.n_clusters, data.shape[1], self.random_state, data.device)
        new_centroids = centroids.copy()

        for epoch in range(self.max_iter):
            # calculate the distance matrix and determine the closest centroid
            distances = ((data - centroids) ** 2).sum(axis=1, keepdim=True)
            matching_centroids = distances.argmin(axis=2, keepdim=True)

            # update the centroids
            for i in range(self.n_clusters):
                selection = (matching_centroids == i).astype(ht.int64)
                new_centroids[:, :, i:i + 1] = ((data * selection).sum(axis=0, keepdim=True) /
                                                selection.sum(axis=0, keepdim=True).clip(1.0, sys.maxsize))

            # check whether centroid movement has converged
            epsilon = ((centroids - new_centroids) ** 2).sum()
            centroids = new_centroids.copy()
            if self.tol is not None and epsilon <= self.tol:
                break

        return centroids



import numpy as np
import torch

from .communication import Communication, MPI, MPI_WORLD
from .stride_tricks import sanitize_axis, sanitize_shape
from . import devices
from . import dndarray
from . import types

__all__ = [
    'arange',
    'array',
    'empty',
    'empty_like',
    'eye',
    'full',
    'full_like',
    'linspace',
    'ones',
    'ones_like',
    'zeros',
    'zeros_like'
]


def arange(*args, dtype=None, split=None, device=None, comm=MPI_WORLD):
    """
    Return evenly spaced values within a given interval.

    Values are generated within the half-open interval ``[start, stop)`` (in other words, the interval including `start`
    but excluding `stop`). For integer arguments the function is equivalent to the Python built-in `range
    <http://docs.python.org/lib/built-in-funcs.html>`_ function, but returns a tensor rather than a list.

    When using a non-integer step, such as 0.1, the results will often not be consistent. It is better to use
    ``linspace`` for these cases.

    Parameters
    ----------
    start : number, optional
        Start of interval.  The interval includes this value.  The default start value is 0.
    stop : number
        End of interval.  The interval does not include this value, except in some cases where `step` is not an integer
        and floating point round-off affects the length of `out`.
    step : number, optional
        Spacing between values.  For any output `out`, this is the distance between two adjacent values, ``out[i+1] -
        out[i]``. The default step size is 1. If `step` is specified as a position argument, `start` must also be given.
    dtype : dtype
        The type of the output array.  If `dtype` is not given, infer the data type from the other input arguments.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    arange : ht.DNDarraytensor
        1D heat tensor of evenly spaced values.

        For floating point arguments, the length of the result is ``ceil((stop - start)/step)``. Because of floating
        point overflow, this rule may result in the last element of `out` being greater than `stop`.

    See Also
    --------
    linspace : Evenly spaced numbers with careful handling of endpoints.

    Examples
    --------
    >>> ht.arange(3)
    tensor([0, 1, 2])
    >>> ht.arange(3.0)
    tensor([ 0.,  1.,  2.])
    >>> ht.arange(3, 7)
    tensor([3, 4, 5, 6])
    >>> ht.arange(3, 7, 2)
    tensor([3, 5])
    """
    num_of_param = len(args)

    # check if all positional arguments are integers
    all_ints = all([isinstance(_, int) for _ in args])

    # set start, stop, step, num according to *args
    if num_of_param == 1:
        if dtype is None:
            # use int32 as default instead of int64 used in numpy
            dtype = types.int32
        start = 0
        stop = int(np.ceil(args[0]))
        step = 1
        num = stop
    elif num_of_param == 2:
        if dtype is None:
            dtype = types.int32 if all_ints else types.float32
        start = args[0]
        stop = args[1]
        step = 1
        num = int(np.ceil(stop - start))
    elif num_of_param == 3:
        if dtype is None:
            dtype = types.int32 if all_ints else types.float32
        start = args[0]
        stop = args[1]
        step = args[2]
        num = int(np.ceil((stop - start) / step))
    else:
        raise TypeError('function takes minimum one and at most 3 positional arguments ({} given)'.format(num_of_param))

    gshape = (num,)
    split = sanitize_axis(gshape, split)
    offset, lshape, _ = comm.chunk(gshape, split)

    # compose the local tensor
    start += offset * step
    stop = start + lshape[0] * step
    device = devices.sanitize_device(device)
    data = torch.arange(
        start, stop, step,
        device=device.torch_device
    )

    htype = types.canonical_heat_type(dtype)
    data = data.type(htype.torch_type())

    return dndarray.DNDarray(data, gshape, htype, split, device, comm)


def array(obj, dtype=None, copy=True, ndmin=0, split=None, is_split=None, device=None, comm=MPI_WORLD):
    """
    Create a tensor.
    Parameters
    ----------
    obj : array_like
        A tensor or array, any object exposing the array interface, an object whose __array__ method returns an array,
        or any (nested) sequence.
    dtype : dtype, optional
        The desired data-type for the array. If not given, then the type will be determined as the minimum type required
        to hold the objects in the sequence. This argument can only be used to ‘upcast’ the array. For downcasting, use
        the .astype(t) method.
    copy : bool, optional
        If true (default), then the object is copied. Otherwise, a copy will only be made if obj is a nested sequence or
        if a copy is needed to satisfy any of the other requirements, e.g. dtype.
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting array should have. Ones will, if needed, be attached to the
        shape if  ndim>0  and prefaced in case of ndim<0 to meet the requirement.
    split : None or int, optional
        The axis along which the passed array content obj is split and distributed in memory. Mutually exclusive with
        is_split.
    is_split : None or int, optional
        Specifies the axis along which the local data portions, passed in obj, are split across all machines. Useful for
        interfacing with other HPC code. The shape of the global tensor is automatically inferred. Mutually exclusive
        with split.
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed tensor chunks.

    Returns
    -------
    out : ht.DNDarray
        A tensor object satisfying the specified requirements.

    Examples
    --------
    >>> ht.array([1, 2, 3])
    tensor([1, 2, 3])

    Upcasting:
    >>> ht.array([1, 2, 3.0])
    tensor([ 1.,  2.,  3.])

    More than one dimension:
    >>> ht.array([[1, 2], [3, 4]])
    tensor([[1, 2],
            [3, 4]])

    Minimum dimensions given:
    >>> ht.array([1, 2, 3], ndmin=2)
    tensor([[1, 2, 3]])

    Type provided:
    >>> ht.array([1, 2, 3], dtype=float)
    tensor([ 1.0, 2.0, 3.0])

    Split data:
    >>> ht.array([1, 2, 3, 4], split=0)
    (0/2) tensor([1, 2])
    (1/2) tensor([3, 4])

    Pre-split data:
    (0/2) >>> ht.array([1, 2], is_split=0)
    (1/2) >>> ht.array([3, 4], is_split=0)
    (0/2) tensor([1, 2, 3, 4])
    (1/2) tensor([1, 2, 3, 4])
    """
    # extract the internal tensor in case of a heat tensor
    if isinstance(obj, dndarray.DNDarray):
        obj = obj._DNDarray__array

    # sanitize the data type
    if dtype is not None:
        dtype = types.canonical_heat_type(dtype)

    # initialize the array
    if bool(copy):
        if isinstance(obj, torch.Tensor):
            obj = obj.clone().detach()
        else:
            try:
                obj = torch.tensor(obj, dtype=dtype.torch_type() if dtype is not None else None)
            except RuntimeError:
                raise TypeError('invalid data of type {}'.format(type(obj)))

    # infer dtype from obj if not explicitly given
    if dtype is None:
        dtype = types.canonical_heat_type(obj.dtype)

    # sanitize minimum number of dimensions
    if not isinstance(ndmin, int):
        raise TypeError('expected ndmin to be int, but was {}'.format(type(ndmin)))

    # reshape the object to encompass additional dimensions
    ndmin_abs = abs(ndmin) - len(obj.shape)
    if ndmin_abs > 0 and ndmin>0:
        obj = obj.reshape(obj.shape + ndmin_abs * (1,))
    if ndmin_abs > 0 and ndmin<0:
        obj = obj.reshape(ndmin_abs * (1,) + obj.shape )

    # sanitize the split axes, ensure mutual exclusiveness
    split = sanitize_axis(obj.shape, split)
    is_split = sanitize_axis(obj.shape, is_split)
    if split is not None and is_split is not None:
        raise ValueError('split and is_split are mutually exclusive parameters')

    # sanitize communication object
    if not isinstance(comm, Communication):
        raise TypeError('expected communication object, but got {}'.format(type(comm)))

    # determine the local and the global shape, if not split is given, they are identical
    lshape = np.array(obj.shape)
    gshape = lshape.copy()

    # content shall be split, chunk the passed data object up
    if split is not None:
        _, _, slices = comm.chunk(obj.shape, split)
        obj = obj[slices]
    # check with the neighboring rank whether the local shape would fit into a global shape
    elif is_split is not None:
        if comm.rank < comm.size - 1:
            comm.Isend(lshape, dest=comm.rank + 1)
        if comm.rank != 0:
            # look into the message of the neighbor to see whether the shape length fits
            status = MPI.Status()
            comm.Probe(source=comm.rank - 1, status=status)
            length = status.Get_count() // lshape.dtype.itemsize

            # the number of shape elements does not match with the 'left' rank
            if length != len(lshape):
                discard_buffer = np.empty(length)
                comm.Recv(discard_buffer, source=comm.rank - 1)
                gshape[is_split] = np.iinfo(gshape.dtype).min
            else:
                # check whether the individual shape elements match
                comm.Recv(gshape, source=comm.rank - 1)
                for i in range(length):
                    if i == is_split:
                        continue
                    elif lshape[i] != gshape[i] and lshape[i] - 1 != gshape[i]:
                        gshape[is_split] = np.iinfo(gshape.dtype).min

        # sum up the elements along the split dimension
        reduction_buffer = np.array(gshape[is_split])
        comm.Allreduce(MPI.IN_PLACE, reduction_buffer, MPI.SUM)
        if reduction_buffer < 0:
            raise ValueError('unable to construct tensor, shape of local data chunk does not match')
        gshape[is_split] = reduction_buffer
        split = is_split

    return dndarray.DNDarray(obj, tuple(int(ele) for ele in gshape), dtype, split, device, comm)


def empty(shape, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new uninitialized array of given shape and data type. May be allocated split up across multiple
    nodes along the specified axis.

    Parameters
    ----------
    shape : int or sequence of ints
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : ht.dtype
        The desired HeAT data type for the array, defaults to ht.float32.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of zeros with given shape, data type and node distribution.

    Examples
    --------
    >>> ht.empty(3)
    tensor([ 0.0000e+00, -2.0000e+00,  3.3113e+35])

    >>> ht.empty(3, dtype=ht.int)
    tensor([ 0.0000e+00, -2.0000e+00,  3.3113e+35])

    >>> ht.empty((2, 3,))
    tensor([[ 0.0000e+00, -2.0000e+00,  3.3113e+35],
            [ 3.6902e+19,  1.2096e+04,  7.1846e+22]])
    """
    return __factory(shape, dtype, split, torch.empty, device, comm)


def empty_like(a, dtype=None, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new uninitialized array with the same type, shape and data distribution of given object. Data type and
    data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : object
        The shape and data-type of 'a' define these same attributes of the returned array.
        Uninitialized tensor with the same shape, type and split axis as 'a' unless overriden.
    dtype : ht.dtype, optional
        Overrides the data type of the result.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        A new uninitialized array.

    Examples
    --------
    >>> x = ht.ones((2, 3,))
    >>> x
    tensor([[1., 1., 1.],
            [1., 1., 1.]])

    >>> ht.empty_like(x)
    tensor([[ 0.0000e+00, -2.0000e+00,  3.3113e+35],
            [ 3.6902e+19,  1.2096e+04,  7.1846e+22]])
    """
    return __factory_like(a, dtype, split, empty, device, comm)


def eye(shape, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new 2-D tensor with ones on the diagonal and zeroes elsewhere.

    Parameters
    ----------
    shape : int or tuple of ints
            The shape of the data-type. If only one number is provided, returning tensor will be square with that size.
            In other cases, the first value represents the number rows, the second the number of columns.
    dtype : ht.dtype, optional
            Overrides the data type of the result.
    split : int, optional
            The axis along which the tensor is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
            Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm : Communication, optional
            Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        An identity matrix.

    Examples
    --------
    >>> import heat as ht
    >>> ht.eye(2)
    tensor([[1., 0.],
            [0., 1.]])

    >>> ht.eye((2, 3), dtype=ht.int32)
    tensor([[1, 0, 0],
            [0, 1, 0]], dtype=torch.int32)
    """
    # Determine the actual size of the resulting data
    gshape = shape
    if isinstance(gshape, int):
        gshape = (gshape, gshape)
    if len(gshape) is 1:
        gshape = gshape * 2

    split = sanitize_axis(gshape, split)
    device = devices.sanitize_device(device)
    offset, lshape, _ = comm.chunk(gshape, split)
    # Start by creating tensor filled with zeroes
    data = torch.zeros(lshape, dtype=types.canonical_heat_type(dtype).torch_type(), device=device.torch_device)
    # Insert ones at the correct positions
    for i in range(min(lshape)):
        pos_x = i if split is 0 else i + offset
        pos_y = i if split is 1 else i + offset
        data[pos_x][pos_y] = 1

    return dndarray.DNDarray(data, gshape, types.canonical_heat_type(data.dtype), split, device, comm)


def __factory(shape, dtype, split, local_factory, device, comm):
    """
    Abstracted factory function for HeAT tensor initialization.

    Parameters
    ----------
    shape : int or sequence of ints
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : ht.dtype
        The desired HeAT data type for the array, defaults to ht.float32.
    split : int
        The axis along which the array is split and distributed.
    local_factory : function
        Function that creates the local PyTorch tensor for the HeAT tensor.
    device : str or None
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of ones with given shape, data type and node distribution.
    """
    # clean the user input
    shape = sanitize_shape(shape)
    dtype = types.canonical_heat_type(dtype)
    split = sanitize_axis(shape, split)
    device = devices.sanitize_device(device)

    # chunk the shape if necessary
    _, local_shape, _ = comm.chunk(shape, split)
    # create the torch data using the factory function
    data = local_factory(local_shape, dtype=dtype.torch_type(), device=device.torch_device)

    return dndarray.DNDarray(data, shape, dtype, split, device, comm)


def __factory_like(a, dtype, split, factory, device, comm, **kwargs):
    """
    Abstracted '...-like' factory function for HeAT tensor initialization

    Parameters
    ----------
    a : object
        The shape and data-type of 'a' define these same attributes of the returned array.
    dtype : ht.dtype
        The desired HeAT data type for the array, defaults to ht.float32.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    factory : function
        Function that creates a HeAT tensor.
    device : str or None
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of ones with given shape, data type and node distribution that is like a
    """
    # determine the global shape of the object to create
    # attempt in this order: shape property, length of object or default shape (1,)
    try:
        shape = a.shape
    except AttributeError:
        try:
            shape = (len(a),)
        except TypeError:
            shape = (1,)

    # infer the data type, otherwise default to float32
    if dtype is None:
        try:
            dtype = types.heat_type_of(a)
        except TypeError:
            dtype = types.float32

    # infer split axis
    if split is None:
        try:
            split = a.split if not isinstance(a, str) else None
        except AttributeError:
            # do not split at all
            pass

    return factory(shape, dtype=dtype, split=split, device=device, comm=comm, **kwargs)


def full(shape, fill_value, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Return a new array of given shape and type, filled with fill_value.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new array, e.g., (2, 3) or 2.
    fill_value : scalar
        Fill value.
    dtype : data-type, optional
        The desired data-type for the array
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of fill_value with the given shape, dtype and split.

    Examples
    --------
    >>> ht.full((2, 2), np.inf)
    tensor([[ inf,  inf],
            [ inf,  inf]])
    >>> ht.full((2, 2), 10)
    tensor([[10, 10],
            [10, 10]])
    """
    def local_factory(*args, **kwargs):
        return torch.full(*args, fill_value=fill_value, **kwargs)

    return __factory(shape, dtype, split, local_factory, device, comm)


def full_like(a, fill_value, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Return a full array with the same shape and type as a given array.

    Parameters
    ----------
    a : object
        The shape and data-type of 'a' define these same attributes of the returned array.
    fill_value : scalar
        Fill value.
    dtype : ht.dtype, optional
        Overrides the data type of the result.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of fill_value with the same shape and type as a.

    Examples
    --------
    >>> x = ht.zeros((2, 3,))
    >>> x
    tensor([[0., 0., 0.],
            [0., 0., 0.]])

    >>> ht.full_like(a, 1.0)
    tensor([[1., 1., 1.],
            [1., 1., 1.]])
    """
    return __factory_like(a, dtype, split, full, device, comm, fill_value=fill_value)


def linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=None, split=None, device=None, comm=MPI_WORLD):
    """
    Returns num evenly spaced samples, calculated over the interval [start, stop]. The endpoint of the interval can
    optionally be excluded.

    Parameters
    ----------
    start: scalar, scalar-convertible
        The starting value of the sample interval, maybe a sequence if convertible to scalar
    end: scalar, scalar-convertible
        The end value of the sample interval, unless is set to False. In that case, the sequence consists of all but the
        last of num + 1 evenly spaced samples, so that stop is excluded. Note that the step size changes when endpoint
        is False.
    num: int, optional
        Number of samples to generate, defaults to 50. Must be non-negative.
    endpoint: bool, optional
        If True, stop is the last sample, otherwise, it is not included. Defaults to True.
    retstep: bool, optional
        If True, return (samples, step), where step is the spacing between samples.
    dtype: dtype, optional
        The type of the output array.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    samples: ht.DNDarray
        There are num equally spaced samples in the closed interval [start, stop] or the half-open interval
        [start, stop) (depending on whether endpoint is True or False).
    step: float, optional
        Size of spacing between samples, only returned if retstep is True.

    Examples
    --------
    >>> ht.linspace(2.0, 3.0, num=5)
    tensor([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ])
    >>> ht.linspace(2.0, 3.0, num=5, endpoint=False)
    tensor([ 2. ,  2.2,  2.4,  2.6,  2.8])
    >>> ht.linspace(2.0, 3.0, num=5, retstep=True)
    (array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ]), 0.25)
    """
    # sanitize input parameters
    start = float(start)
    stop = float(stop)
    num = int(num)
    if num <= 0:
        raise ValueError('number of samples \'num\' must be non-negative integer, but was {}'.format(num))
    step = (stop - start) / max(1, num - 1 if endpoint else num)

    # infer local and global shapes
    gshape = (num,)
    split = sanitize_axis(gshape, split)
    offset, lshape, _ = comm.chunk(gshape, split)

    # compose the local tensor
    start += offset * step
    stop = start + lshape[0] * step - step
    device = devices.sanitize_device(device)
    data = torch.linspace(start, stop, lshape[0], device=device.torch_device)
    if dtype is not None:
        data = data.type(types.canonical_heat_type(dtype).torch_type())

    # construct the resulting global tensor
    ht_tensor = dndarray.DNDarray(data, gshape, types.canonical_heat_type(data.dtype), split, device, comm)

    if retstep:
        return ht_tensor, step
    return ht_tensor


def ones(shape, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new array of given shape and data type filled with one values. May be allocated split up across multiple
    nodes along the specified axis.

    Parameters
    ----------
    shape : int or sequence of ints
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : ht.dtype
        The desired HeAT data type for the array, defaults to ht.float32.
    split : int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm : Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of ones with given shape, data type and node distribution.

    Examples
    --------
    >>> ht.ones(3)
    tensor([1., 1., 1.])

    >>> ht.ones(3, dtype=ht.int)
    tensor([1, 1, 1])

    >>> ht.ones((2, 3,))
    tensor([[1., 1., 1.],
            [1., 1., 1.]])
    """
    return __factory(shape, dtype, split, torch.ones, device, comm)


def ones_like(a, dtype=None, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new array filled with ones with the same type, shape and data distribution of given object. Data type and
    data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : object
        The shape and data-type of 'a' define these same attributes of the returned array.
    dtype : ht.dtype, optional
        Overrides the data type of the result.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of ones with the same shape, type and split axis as 'a' unless overriden.

    Examples
    --------
    >>> x = ht.zeros((2, 3,))
    >>> x
    tensor([[0., 0., 0.],
            [0., 0., 0.]])

    >>> ht.ones_like(a)
    tensor([[1., 1., 1.],
            [1., 1., 1.]])
    """
    return __factory_like(a, dtype, split, ones, device, comm)


def zeros(shape, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new array of given shape and data type filled with zero values. May be allocated split up across multiple
    nodes along the specified axis.

    Parameters
    ----------
    shape : int or sequence of ints
        Desired shape of the output array, e.g. 1 or (1, 2, 3,).
    dtype : ht.dtype
        The desired HeAT data type for the array, defaults to ht.float32.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of zeros with given shape, data type and node distribution.

    Examples
    --------
    >>> ht.zeros(3)
    tensor([0., 0., 0.])

    >>> ht.zeros(3, dtype=ht.int)
    tensor([0, 0, 0])

    >>> ht.zeros((2, 3,))
    tensor([[0., 0., 0.],
            [0., 0., 0.]])
    """
    return __factory(shape, dtype, split, torch.zeros, device, comm)


def zeros_like(a, dtype=None, split=None, device=None, comm=MPI_WORLD):
    """
    Returns a new array filled with zeros with the same type, shape and data distribution of given object. Data type and
    data distribution strategy can be explicitly overriden.

    Parameters
    ----------
    a : object
        The shape and data-type of 'a' define these same attributes of the returned array.
    dtype : ht.dtype, optional
        Overrides the data type of the result.
    split: int, optional
        The axis along which the array is split and distributed, defaults to None (no distribution).
    device : str, ht.Device or None, optional
        Specifies the device the tensor shall be allocated on, defaults to None (i.e. globally set default device).
    comm: Communication, optional
        Handle to the nodes holding distributed parts or copies of this tensor.

    Returns
    -------
    out : ht.DNDarray
        Array of zeros with the same shape, type and split axis as 'a' unless overriden.

    Examples
    --------
    >>> x = ht.ones((2, 3,))
    >>> x
    tensor([[1., 1., 1.],
            [1., 1., 1.]])

    >>> ht.zeros_like(x)
    tensor([[0., 0., 0.],
            [0., 0., 0.]])
    """
    return __factory_like(a, dtype, split, zeros, device, comm)



"""
types: Defines the basic heat data types in the hierarchy as shown below. Design inspired by the Python package numpy.

As part of the type-hierarchy: xx -- is bit-width
generic
 +-> bool, bool_                                (kind=?)
 +-> number
 |   +-> integer
 |   |   +-> signedinteger         (intxx)      (kind=b, i)
 |   |   |     int8, byte
 |   |   |     int16, short
 |   |   |     int32, int
 |   |   |     int64, long
 |   |   \\-> unsignedinteger      (uintxx)     (kind=B, u)
 |   |         uint8, ubyte
 |   \\-> floating                 (floatxx)    (kind=f)
 |         float32, float, float_
 |         float64, double         (double)
 \\-> flexible (currently unused, placeholder for characters)
"""

import abc
import builtins
import collections
import numpy as np
import torch

from . import devices
from .communication import MPI_WORLD


__all__ = [
    'generic',
    'number',
    'integer',
    'signedinteger',
    'unsignedinteger',
    'bool',
    'bool_',
    'floating',
    'int8',
    'byte',
    'int16',
    'short',
    'int32',
    'int',
    'int64',
    'long',
    'uint8',
    'ubyte',
    'float32',
    'float',
    'float_',
    'float64',
    'double',
    'flexible',
    'can_cast',
    'promote_types'
]


class generic(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __new__(cls, *value, device=None, comm=MPI_WORLD):
        try:
            torch_type = cls.torch_type()
        except TypeError:
            raise TypeError('cannot create \'{}\' instances'.format(cls))

        value_count = len(value)

        # check whether there are too many arguments
        if value_count >= 2:
            raise TypeError('function takes at most 1 argument ({} given)'.format(value_count))
        # if no value is given, we will initialize the value to be zero
        elif value_count == 0:
            value = ((0,),)

        # otherwise, attempt to create a torch tensor of given type
        try:
            array = torch.tensor(*value, dtype=torch_type)
        except TypeError as exception:
            # re-raise the exception to be consistent with numpy's exception interface
            raise ValueError(str(exception))

        # sanitize the input device type
        device = devices.sanitize_device(device)

        return dndarray.DNDarray(array, tuple(array.shape), cls, split=None, device=device, comm=comm)

    @classmethod
    @abc.abstractclassmethod
    def torch_type(cls):
        pass

    @classmethod
    @abc.abstractclassmethod
    def char(cls):
        pass


class bool(generic):
    @classmethod
    def torch_type(cls):
        return torch.uint8

    @classmethod
    def char(cls):
        return 'u1'


class number(generic):
    pass


class integer(number):
    pass


class signedinteger(integer):
    pass


class int8(signedinteger):
    @classmethod
    def torch_type(cls):
        return torch.int8

    @classmethod
    def char(cls):
        return 'i1'


class int16(signedinteger):
    @classmethod
    def torch_type(cls):
        return torch.int16

    @classmethod
    def char(cls):
        return 'i2'


class int32(signedinteger):
    @classmethod
    def torch_type(cls):
        return torch.int32

    @classmethod
    def char(cls):
        return 'i4'


class int64(signedinteger):
    @classmethod
    def torch_type(cls):
        return torch.int64

    @classmethod
    def char(cls):
        return 'i8'


class unsignedinteger(integer):
    pass


class uint8(unsignedinteger):
    @classmethod
    def torch_type(cls):
        return torch.uint8

    @classmethod
    def char(cls):
        return 'u1'


class floating(number):
    pass


class float32(floating):
    @classmethod
    def torch_type(cls):
        return torch.float32

    @classmethod
    def char(cls):
        return 'f4'


class float64(floating):
    @classmethod
    def torch_type(cls):
        return torch.float64

    @classmethod
    def char(cls):
        return 'f8'


class flexible(generic):
    pass


# definition of aliases
bool_ = bool
ubyte = uint8
byte = int8
short = int16
int = int32
int_ = int32
long = int64
float = float32
float_ = float32
double = float64

_inexact = (
    # float16,
    float32,
    float64,)

_exact = (
    uint8,
    int8,
    int16,
    int32,
    int64,)

# type mappings for type strings and builtins types
__type_mappings = {
    # type strings
    '?':            bool,
    'b':            int8,
    'i':            int32,
    'i1':           int8,
    'i2':           int16,
    'i4':           int32,
    'i8':           int64,
    'B':            uint8,
    'u':            uint8,
    'u1':           uint8,
    'f':            float32,
    'f4':           float32,
    'f8':           float64,

    # numpy types
    np.bool:        bool,
    np.uint8:       uint8,
    np.int8:        int8,
    np.int16:       int16,
    np.int32:       int32,
    np.int64:       int64,
    np.float32:     float32,
    np.float64:     float64,

    # torch types
    torch.uint8:    uint8,
    torch.int8:     int8,
    torch.int16:    int16,
    torch.int32:    int32,
    torch.int64:    int64,
    torch.float32:  float32,
    torch.float64:  float64,

    # builtins
    builtins.bool:  bool,
    builtins.int:   int32,
    builtins.float: float32,
}


def canonical_heat_type(a_type):
    """
    Canonicalize the builtin Python type, type string or HeAT type into a canonical HeAT type.

    Parameters
    ----------
    a_type : type, str, ht.dtype
        A description for the type. It may be a a Python builtin type, string or an HeAT type already.
        In the three former cases the according mapped type is looked up, in the latter the type is simply returned.

    Returns
    -------
    out : ht.dtype
        The canonical HeAT type.

    Raises
    -------
    TypeError
        If the type cannot be converted.
    """
    # already a heat type
    try:
        if issubclass(a_type, generic):
            return a_type
    except TypeError:
        pass

    # try to look the corresponding type up
    try:
        return __type_mappings[a_type]
    except KeyError:
        raise TypeError('data type {} is not understood'.format(a_type))


def heat_type_of(obj):
    """
    Returns the corresponding HeAT data type of given object, i.e. scalar, array or iterable. Attempts to determine the
    canonical data type based on the following priority list:
        1. dtype property
        2. type(obj)
        3. type(obj[0])

    Parameters
    ----------
    obj : scalar, array, iterable
        The object for which to infer the type.

    Returns
    -------
    out : ht.dtype
        The object's corresponding HeAT type.

    Raises
    -------
    TypeError
        If the object's type cannot be inferred.
    """
    # attempt to access the dtype property
    try:
        return canonical_heat_type(obj.dtype)
    except (AttributeError, TypeError,):
        pass

    # attempt type of object itself
    try:
        return canonical_heat_type(type(obj))
    except TypeError:
        pass

    # last resort, type of the object at first position
    try:
        return canonical_heat_type(type(obj[0]))
    except (KeyError, IndexError, TypeError,):
        raise TypeError('data type of {} is not understood'.format(obj))


# type code assignment
__type_codes = collections.OrderedDict([
    (bool,    0),
    (uint8,   1),
    (int8,    2),
    (int16,   3),
    (int32,   4),
    (int64,   5),
    (float32, 6),
    (float64, 7),
])

# safe cast table
__safe_cast = [
    # bool  uint8  int8   int16  int32  int64  float32 float64
    [True,  True,  True,  True,  True,  True,  True,   True],  # bool
    [False, True,  False, True,  True,  True,  True,   True],  # uint8
    [False, False, True,  True,  True,  True,  True,   True],  # int8
    [False, False, False, True,  True,  True,  True,   True],  # int16
    [False, False, False, False, True,  True,  False,  True],  # int32
    [False, False, False, False, False, True,  False,  True],  # int64
    [False, False, False, False, False, False, True,   True],  # float32
    [False, False, False, False, False, False, False,  True]   # float64
]


# same kind table
__same_kind = [
    # bool  uint8  int8   int16  int32  int64  float32 float64
    [True,  False, False, False, False, False, False,  False],  # bool
    [False, True,  True,  True,  True,  True,  False,  False],  # uint8
    [False, True,  True,  True,  True,  True,  False,  False],  # int8
    [False, True,  True,  True,  True,  True,  False,  False],  # int16
    [False, True,  True,  True,  True,  True,  False,  False],  # int32
    [False, True,  True,  True,  True,  True,  False,  False],  # int64
    [False, False, False, False, False, False, True,   True],   # float32
    [False, False, False, False, False, False, True,   True]    # float64
]


# static list of possible casting methods
__cast_kinds = ['no', 'safe', 'same_kind', 'unsafe']


def can_cast(from_, to, casting='safe'):
    """
    Returns True if cast between data types can occur according to the casting rule. If from is a scalar or array
    scalar, also returns True if the scalar value can be cast without overflow or truncation to an integer.

    Parameters
    ----------
    from_ : scalar, tensor, type, str, ht.dtype
        Scalar, data type or type specifier to cast from.
    to : type, str, ht.dtype
        Target type to cast to.
    casting: str {'no', 'safe', 'same_kind', 'unsafe'}, optional
        Controls the way the cast is evaluated
            * 'no' the types may not be cast, i.e. they need to be identical
            * 'safe' allows only casts that can preserve values with complete precision
            * 'same_kind' safe casts are possible and down_casts within the same type family, e.g. int32 -> int8
            * 'unsafe' means any conversion can be performed, i.e. this casting is always possible

    Returns
    -------
    out : bool
        True if cast can occur according to the casting rules, False otherwise.

    Raises
    -------
    TypeError
        If the types are not understood or casting is not a string
    ValueError
        If the casting rule is not understood

    Examples
    --------
    Basic examples with types

    >>> ht.can_cast(ht.int32, ht.int64)
    True
    >>> ht.can_cast(ht.int64, ht.float64)
    True
    >>> ht.can_cast(ht.int16, ht.int8)
    False

    The usage of scalars is also possible
    >>> ht.can_cast(1, ht.float64)
    True
    >>> ht.can_cast(2.0e200, 'u1')
    False

    can_cast supports different casting rules
    >>> ht.can_cast('i8', 'i4', 'no')
    False
    >>> ht.can_cast('i8', 'i4', 'safe')
    False
    >>> ht.can_cast('i8', 'i4', 'same_kind')
    True
    >>> ht.can_cast('i8', 'i4', 'unsafe')
    True
    """
    if not isinstance(casting, str):
        raise TypeError('expected string, found {}'.format(type(casting)))
    if casting not in __cast_kinds:
        raise ValueError('casting must be one of {}'.format(str(__cast_kinds)[1:-1]))

    # obtain the types codes of the canonical HeAT types
    try:
        typecode_from = __type_codes[canonical_heat_type(from_)]
    except TypeError:
        typecode_from = __type_codes[heat_type_of(from_)]
    typecode_to = __type_codes[canonical_heat_type(to)]

    # unsafe casting allows everything
    if casting == 'unsafe':
        return True

    # types have to match exactly
    elif casting == 'no':
        return typecode_from == typecode_to

    # safe casting or same_kind
    can_safe_cast = __safe_cast[typecode_from][typecode_to]
    if casting == 'safe':
        return can_safe_cast
    return can_safe_cast or __same_kind[typecode_from][typecode_to]


# compute possible type promotions dynamically
__type_promotions = [[None] * len(row) for row in __same_kind]
for i, operand_a in enumerate(__type_codes.keys()):
    for j, operand_b in enumerate(__type_codes.keys()):
        for target in __type_codes.keys():
            if can_cast(operand_a, target) and can_cast(operand_b, target):
                __type_promotions[i][j] = target
                break


def promote_types(type1, type2):
    """
    Returns the data type with the smallest size and smallest scalar kind to which both type1 and type2 may be safely
    cast. This function is symmetric.

    Parameters
    ----------
    type1 : type, str, ht.dtype
        type of first operand
    type2 : type, str, ht.dtype
        type of second operand

    Returns
    -------
    out : ht.dtype
        The promoted data type.

    Examples
    --------
    >>> ht.promote_types(ht.uint8, ht.uint8)
    ht.uint8
    >>> ht.promote_types(ht.int8, ht.uint8)
    ht.int16
    >>> ht.promote_types('i8', 'f4')
    ht.float64
    """
    typecode_type1 = __type_codes[canonical_heat_type(type1)]
    typecode_type2 = __type_codes[canonical_heat_type(type2)]

    return __type_promotions[typecode_type1][typecode_type2]


class finfo:
    """
    finfo(dtype)

    Class describing machine limits (bit representation) of floating point types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    eps : float
        The smallest representable positive number such that
        ``1.0 + eps != 1.0``.  Type of `eps` is an appropriate floating
        point type.
    max : floating point number of the appropriate type
        The largest representable number.
    min : floating point number of the appropriate type
        The smallest representable number, typically ``-max``.
    tiny : float
        The smallest positive usable number.  Type of `tiny` is an
        appropriate floating point type.

    Parameters
    ----------
    dtype : ht.dtype
        Kind of floating point data-type about which to get information.

    Examples:
    ---------
    >>> import heat as ht
    >>> info = ht.types.finfo(ht.float32)
    >>> info.bits
    32

    >>> info.eps
    1.1920928955078125e-07

    """
    def __new__(cls, dtype):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError,):
            # If given type is not heat type
            pass

        if not dtype in _inexact:
            raise TypeError('Data type {} not inexact, not supported'.format(dtype))

        return super(finfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype):
        _torch_finfo = torch.finfo(dtype.torch_type())
        for word in ['bits', 'eps', 'max', 'tiny']:
            setattr(self, word, getattr(_torch_finfo, word))

        self.min = -self.max

        return self


class iinfo:
    """
    iinfo(dtype)

    Class describing machine limits (bit representation) of integer types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    max : floating point number of the appropriate type
        The largest representable number.
    min : floating point number of the appropriate type
        The smallest representable number, typically ``-max``.

    Parameters
    ----------
    dtype : ht.dtype
        Kind of floating point data-type about which to get information.

    Examples:
    ---------
    >>> import heat as ht
    >>> info = ht.types.finfo(ht.int32)
    >>> info.bits
    32

    """
    def __new__(cls, dtype):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError,):
            # If given type is not heat type
            pass

        if not dtype in _exact:
            raise TypeError('Data type {} not inexact, not supported'.format(dtype))

        return super(iinfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype):
        _torch_iinfo = torch.iinfo(dtype.torch_type())
        for word in ['bits', 'max']:
            setattr(self, word, getattr(_torch_iinfo, word))

        self.min = -(self.max + 1)

        return self


# tensor is imported at the very end to break circular dependency
from . import dndarray



# Definition of HeAT Constants

__all__ = [
    'e',
    'Euler',
    'inf',
    'Inf',
    'Infty',
    'Infinity',
    'nan',
    'NaN',
    'pi',
]

INF = float('inf')
NAN = float('nan')
NINF = - float('inf')
PI = 3.141592653589793
E = 2.718281828459045

# aliases
inf = Inf = Infty = Infinity = INF
nan = NaN = NAN
pi = PI
e = Euler = E



import itertools
import torch

from . import dndarray

__all__ = [
    'transpose',
    'tril',
    'triu'
]


def transpose(a, axes=None):
    """
    Permute the dimensions of an array.

    Parameters
    ----------
    a : array_like
        Input array.
    axes : None or list of ints, optional
        By default, reverse the dimensions, otherwise permute the axes according to the values given.

    Returns
    -------
    p : ht.DNDarray
        a with its axes permuted.
    """
    # type check the input tensor
    if not isinstance(a, dndarray.DNDarray):
        raise TypeError('a must be of type ht.DNDarray, but was {}'.format(type(a)))

    # set default value for axes permutations
    dimensions = len(a.shape)
    if axes is None:
        axes = tuple(reversed(range(dimensions)))
    # if given, sanitize the input
    else:
        try:
            # convert to a list to allow index access
            axes = list(axes)
        except TypeError:
            raise ValueError('axes must be an iterable containing ints')

        if len(axes) != dimensions:
            raise ValueError('axes do not match tensor shape')
        for index, axis in enumerate(axes):
            if not isinstance(axis, int):
                raise TypeError('axis must be an integer, but was {}'.format(type(axis)))
            elif axis < 0:
                axes[index] = axis + dimensions

    # infer the new split axis, it is the position of the split axis within the new axes permutation
    try:
        transposed_split = axes.index(a.split) if a.split is not None else None
    except ValueError:
        raise ValueError('axes do not match tensor shape')

    # try to rearrange the tensor and return a new transposed variant
    try:
        transposed_data = a._DNDarray__array.permute(*axes)
        transposed_shape = tuple(a.shape[axis] for axis in axes)

        return dndarray.DNDarray(transposed_data, transposed_shape, a.dtype, transposed_split, a.device, a.comm)
    # if not possible re- raise any torch exception as ValueError
    except (RuntimeError, IndexError) as exception: 
        raise ValueError(str(exception))


# statically allocated index slices for non-iterable dimensions in triangular operations
__index_base = (slice(None), slice(None),)


def __tri_op(m, k, op):
    """
    Generic implementation of triangle operations on tensors. It takes care of input sanitation and non-standard
    broadcast behavior of the 2D triangle-operators.

    Parameters
    ----------
    m : ht.DNDarray
        Input tensor for which to compute the triangle operator.
    k : int, optional
        Diagonal above which to apply the triangle operator, k<0 is below and k>0 is above.
    op : callable
        Implementation of the triangle operator.

    Returns
    -------
    triangle_tensor : ht.DNDarray
        DNDarray with the applied triangle operation

    Raises
    ------
    TypeError
        If the input is not a tensor or the diagonal offset cannot be converted to an integral value.
    """
    if not isinstance(m, dndarray.DNDarray):
        raise TypeError('Expected m to be a tensor but was {}'.format(type(m)))

    try:
        k = int(k)
    except ValueError:
        raise TypeError('Expected k to be integral, but was {}'.format(type(k)))

    # chunk the global shape of the tensor to obtain the offset compared to the other ranks
    offset, _, _ = m.comm.chunk(m.shape, m.split)
    dimensions = len(m.shape)

    # manually repeat the input for vectors
    if dimensions == 1:
        triangle = m._DNDarray__array.expand(m.shape[0], -1)
        if torch.numel(triangle > 0):
            triangle = op(triangle, k - offset)

        return dndarray.DNDarray(
            triangle,
            (m.shape[0], m.shape[0],),
            m.dtype,
            None if m.split is None else 1,
            m.device,
            m.comm
        )

    original = m._DNDarray__array
    output = original.clone()

    # modify k to account for tensor splits
    if m.split is not None:
        if m.split + 1 == dimensions - 1:
            k += offset
        elif m.split == dimensions - 1:
            k -= offset

    # in case of two dimensions we can just forward the call to the callable
    if dimensions == 2:
        if torch.numel(original) > 0:
            op(original, k, out=output)
    # more than two dimensions: iterate over all but the last two to realize 2D broadcasting
    else:
        ranges = [range(elements) for elements in m.lshape[:-2]]
        for partial_index in itertools.product(*ranges):
            index = partial_index + __index_base
            op(original[index], k, out=output[index])

    return dndarray.DNDarray(output, m.shape, m.dtype, m.split, m.device, m.comm)


def tril(m, k=0):
    """
    Returns the lower triangular part of the tensor, the other elements of the result tensor are set to 0.

    The lower triangular part of the tensor is defined as the elements on and below the diagonal.

    The argument k controls which diagonal to consider. If k=0, all elements on and below the main diagonal are
    retained. A positive value includes just as many diagonals above the main diagonal, and similarly a negative
    value excludes just as many diagonals below the main diagonal.

    Parameters
    ----------
    m : ht.DNDarray
        Input tensor for which to compute the lower triangle.
    k : int, optional
        Diagonal above which to zero elements. k=0 (default) is the main diagonal, k<0 is below and k>0 is above.

    Returns
    -------
    lower_triangle : ht.DNDarray
        Lower triangle of the input tensor.
    """
    return __tri_op(m, k, torch.tril)


def triu(m, k=0):
    """
    Returns the upper triangular part of the tensor, the other elements of the result tensor are set to 0.

    The upper triangular part of the tensor is defined as the elements on and below the diagonal.

    The argument k controls which diagonal to consider. If k=0, all elements on and below the main diagonal are
    retained. A positive value includes just as many diagonals above the main diagonal, and similarly a negative
    value excludes just as many diagonals below the main diagonal.

    Parameters
    ----------
    m : ht.DNDarray
        Input tensor for which to compute the upper triangle.
    k : int, optional
        Diagonal above which to zero elements. k=0 (default) is the main diagonal, k<0 is below and k>0 is above.

    Returns
    -------
    upper_triangle : ht.DNDarray
        Upper triangle of the input tensor.
    """
    return __tri_op(m, k, torch.triu)



import numpy as np
import torch
import warnings

from .communication import MPI
from . import exponential
from . import factories
from . import operations
from . import dndarray
from . import types


__all__ = [
    'argmax',
    'argmin',
    'max',
    'mean',
    'min',
    'std',
    'var'
]


def argmax(x, axis=None, out=None, **kwargs):
    """
    Returns the indices of the maximum values along an axis.

    Parameters
    ----------
    x : ht.DNDarray
        Input array.
    axis : int, optional
        By default, the index is into the flattened tensor, otherwise along the specified axis.
    out : ht.DNDarray, optional.
        If provided, the result will be inserted into this tensor. It should be of the appropriate shape and dtype.

    Returns
    -------
    index_tensor : ht.DNDarray of ints
        Array of indices into the array. It has the same shape as x.shape with the dimension along axis removed.

    Examples
    --------
    >>> import heat as ht
    >>> import torch
    >>> torch.manual_seed(1)
    >>> a = ht.random.randn(3,3)
    >>> a
    tensor([[-0.5631, -0.8923, -0.0583],
            [-0.1955, -0.9656,  0.4224],
            [ 0.2673, -0.4212, -0.5107]])
    >>> ht.argmax(a)
    tensor([5])
    >>> ht.argmax(a, axis=0)
    tensor([[2, 2, 1]])
    >>> ht.argmax(a, axis=1)
    tensor([[2],
            [2],
            [0]])
    """
    def local_argmax(*args, **kwargs):
        axis = kwargs.get('dim', -1)
        shape = x.shape

        # case where the argmin axis is set to None
        # argmin will be the flattened index, computed standalone and the actual minimum value obtain separately
        if len(args) <= 1 and axis < 0:
            indices = torch.argmax(*args, **kwargs).reshape(1)
            maxima = args[0].flatten()[indices]

            # artificially flatten the input tensor shape to correct the offset computation
            axis = x.split
            shape = [np.prod(shape)]
        # usual case where indices and maximum values are both returned. Axis is not equal to None
        else:
            maxima, indices = torch.max(*args, **kwargs)

        # add offset of data chunks if reduction is computed across split axis
        if axis == x.split:
            offset, _, _ = x.comm.chunk(shape, x.split)
            indices += offset

        return torch.cat([maxima.double(), indices.double()])

    # axis sanitation
    if axis is not None and not isinstance(axis, int):
        raise TypeError('axis must be None or int, but was {}'.format(type(axis)))

    # perform the global reduction
    reduced_result = operations.__reduce_op(x, local_argmax, MPI_ARGMAX, axis=axis, out=None, **kwargs)

    # correct the tensor
    reduced_result._DNDarray__array = reduced_result._DNDarray__array.chunk(2)[-1].type(torch.int64)
    reduced_result._DNDarray__dtype = types.int64

    # address lshape/gshape mismatch when axis is 0
    if axis is not None:
        if isinstance(axis, int):
            axis = (axis,)
        if 0 in axis:
            reduced_result._DNDarray__gshape = (1,) + reduced_result._DNDarray__gshape
            if not kwargs.get('keepdim'):
                reduced_result = reduced_result.squeeze(axis=0)

    # set out parameter correctly, i.e. set the storage correctly
    if out is not None:
        if out.shape != reduced_result.shape:
            raise ValueError('Expecting output buffer of shape {}, got {}'.format(reduced_result.shape, out.shape))
        out._DNDarray__array.storage().copy_(reduced_result._DNDarray__array.storage())
        out._DNDarray__array = out._DNDarray__array.type(torch.int64)
        out._DNDarray__dtype = types.int64
        return out

    return reduced_result


def argmin(x, axis=None, out=None, **kwargs):
    """
    Returns the indices of the minimum values along an axis.

    Parameters
    ----------
    x : ht.DNDarray
        Input array.
    axis : int, optional
        By default, the index is into the flattened tensor, otherwise along the specified axis.
    # out : ht.DNDarray, optional. Issue #100
        If provided, the result will be inserted into this tensor. It should be of the appropriate shape and dtype.

    Returns
    -------
    index_tensor : ht.DNDarray of ints
        Array of indices into the array. It has the same shape as x.shape with the dimension along axis removed.

    Examples
    --------
    >>> import heat as ht
    >>> import torch
    >>> torch.manual_seed(1)
    >>> a = ht.random.randn(3,3)
    >>> a
    tensor([[-0.5631, -0.8923, -0.0583],
    [-0.1955, -0.9656,  0.4224],
    [ 0.2673, -0.4212, -0.5107]])
    >>> ht.argmin(a)
    tensor([4])
    >>> ht.argmin(a, axis=0)
    tensor([[0, 1, 2]])
    >>> ht.argmin(a, axis=1)
    tensor([[1],
            [1],
            [2]])
    """
    def local_argmin(*args, **kwargs):
        axis = kwargs.get('dim', -1)
        shape = x.shape

        # case where the argmin axis is set to None
        # argmin will be the flattened index, computed standalone and the actual minimum value obtain separately
        if len(args) <= 1 and axis < 0:
            indices = torch.argmin(*args, **kwargs).reshape(1)
            minimums = args[0].flatten()[indices]

            # artificially flatten the input tensor shape to correct the offset computation
            axis = x.split
            shape = [np.prod(shape)]
        # usual case where indices and minimum values are both returned. Axis is not equal to None
        else:
            minimums, indices = torch.min(*args, **kwargs)

        # add offset of data chunks if reduction is computed across split axis
        if axis == x.split:
            offset, _, _ = x.comm.chunk(shape, x.split)
            indices += offset

        return torch.cat([minimums.double(), indices.double()])

    # axis sanitation
    if axis is not None and not isinstance(axis, int):
        raise TypeError('axis must be None or int, but was {}'.format(type(axis)))

    # perform the global reduction
    reduced_result = operations.__reduce_op(x, local_argmin, MPI_ARGMIN, axis=axis, out=None, **kwargs)

    # correct the tensor
    reduced_result._DNDarray__array = reduced_result._DNDarray__array.chunk(2)[-1].type(torch.int64)
    reduced_result._DNDarray__dtype = types.int64

    # address lshape/gshape mismatch when axis is 0
    if axis is not None:
        if isinstance(axis, int):
            axis = (axis,)
        if 0 in axis:
            reduced_result._DNDarray__gshape = (1,) + reduced_result._DNDarray__gshape
            if not kwargs.get('keepdim'):
                reduced_result = reduced_result.squeeze(axis=0)

    # set out parameter correctly, i.e. set the storage correctly
    if out is not None:
        if out.shape != reduced_result.shape:
            raise ValueError('Expecting output buffer of shape {}, got {}'.format(reduced_result.shape, out.shape))
        out._DNDarray__array.storage().copy_(reduced_result._DNDarray__array.storage())
        out._DNDarray__array = out._DNDarray__array.type(torch.int64)
        out._DNDarray__dtype = types.int64
        return out

    return reduced_result


def max(x, axis=None, out=None, keepdim=None):
    # TODO: initial : scalar, optional Issue #101
    """
    Return the maximum along a given axis.

    Parameters
    ----------
    a : ht.DNDarray
        Input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which to operate. By default, flattened input is used.
        If this is a tuple of ints, the maximum is selected over multiple axes,
        instead of a single axis or all the axes as before.
    out : ht.DNDarray, optional
        Tuple of two output tensors (max, max_indices). Must be of the same shape and buffer length as the expected
        output. The minimum value of an output element. Must be present to allow computation on empty slice.

    Returns
    -------
    maximums : ht.DNDarray
        The maximum along a given axis.

    Examples
    --------
    >>> a = ht.float32([
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
        [10, 11, 12]
    ])
    >>> ht.max(a)
    tensor([12.])
    >>> ht.min(a, axis=0)
    tensor([[10., 11., 12.]])
    >>> ht.min(a, axis=1)
    tensor([[ 3.],
            [ 6.],
            [ 9.],
            [12.]])
    """
    def local_max(*args, **kwargs):
        result = torch.max(*args, **kwargs)
        if isinstance(result, tuple):
            return result[0]
        return result

    return operations.__reduce_op(x, local_max, MPI.MAX, axis=axis, out=out, keepdim=keepdim)


def mean(x, axis=None):
    """
    Calculates and returns the mean of a tensor.
    If a axis is given, the mean will be taken in that direction.

    Parameters
    ----------
    x : ht.DNDarray
        Values for which the mean is calculated for.
    axis : None, Int, iterable, defaults to None
        Axis which the mean is taken in. Default None calculates mean of all data items.

    Returns
    -------
    means : ht.DNDarray
        The mean/s, if split, then split in the same direction as x.

    Examples
    --------
    >>> a = ht.random.randn(1,3)
    >>> a
    tensor([[-1.2435,  1.1813,  0.3509]])
    >>> ht.mean(a)
    tensor(0.0962)

    >>> a = ht.random.randn(4,4)
    >>> a
    tensor([[ 0.0518,  0.9550,  0.3755,  0.3564],
            [ 0.8182,  1.2425,  1.0549, -0.1926],
            [-0.4997, -1.1940, -0.2812,  0.4060],
            [-1.5043,  1.4069,  0.7493, -0.9384]])
    >>> ht.mean(a, 1)
    tensor([ 0.4347,  0.7307, -0.3922, -0.0716])
    >>> ht.mean(a, 0)
    tensor([-0.2835,  0.6026,  0.4746, -0.0921])

    >>> a = ht.random.randn(4,4)
    >>> a
    tensor([[ 2.5893,  1.5934, -0.2870, -0.6637],
            [-0.0344,  0.6412, -0.3619,  0.6516],
            [ 0.2801,  0.6798,  0.3004,  0.3018],
            [ 2.0528, -0.1121, -0.8847,  0.8214]])
    >>> ht.mean(a, (0,1))
    tensor(0.4730)
    """

    def reduce_means_elementwise(output_shape_i):
        """
        Function to combine the calculated means together.
        This does an element-wise update of the calculated means to merge them together using the merge_means function.
        This function operates using x from the mean function paramters.

        Parameters
        ----------
        output_shape_i : iterable
            Iterable with the dimensions of the output of the mean function.

        Returns
        -------
        means : ht.DNDarray
            The calculated means.
        """
        if x.lshape[x.split] != 0:
            mu = torch.mean(x._DNDarray__array, out=None, dim=axis)
        else:
            mu = factories.zeros(output_shape_i)

        mu_shape = list(mu.shape) if list(mu.shape) else [1]

        mu_tot = factories.zeros(([x.comm.size] + mu_shape))
        n_tot = factories.zeros(x.comm.size)
        mu_tot[x.comm.rank, :] = mu
        n_tot[x.comm.rank] = float(x.lshape[x.split])
        x.comm.Allreduce(MPI.IN_PLACE, mu_tot, MPI.SUM)
        x.comm.Allreduce(MPI.IN_PLACE, n_tot, MPI.SUM)

        for i in range(1, x.comm.size):
            mu_tot[0, :],  n_tot[0] = merge_means(mu_tot[0, :], n_tot[0], mu_tot[i, :], n_tot[i])

        return mu_tot[0]

    if axis is None:
        # full matrix calculation
        if not x.is_distributed():
            # if x is not distributed do a torch.mean on x
            ret = torch.mean(x._DNDarray__array)
            return dndarray.DNDarray(ret, tuple(ret.shape), x.dtype, None, x.device, x.comm)
        else:
            # if x is distributed and no axis is given: return mean of the whole set
            if x.lshape[x.split] != 0:
                mu_in = operations.__local_op(torch.mean, x, out=None)
            else:
                mu_in = 0
            n = x.lnumel
            mu_tot = factories.zeros((x.comm.size, 2))
            mu_proc = factories.zeros((x.comm.size, 2))
            mu_proc[x.comm.rank][0] = mu_in
            mu_proc[x.comm.rank][1] = float(n)
            x.comm.Allreduce(mu_proc, mu_tot, MPI.SUM)

            for i in range(1, x.comm.size):
                merged = merge_means(mu_tot[0, 0], mu_tot[0, 1], mu_tot[i, 0], mu_tot[i, 1])
                mu_tot[0, 0] = merged[0]
                mu_tot[0, 1] = merged[1]

            return mu_tot[0][0]
    else:
        output_shape = list(x.shape)
        if isinstance(axis, (list, tuple, dndarray.DNDarray, torch.Tensor)):
            if any([not isinstance(j, int) for j in axis]):
                raise ValueError("items in axis itterable must be integers, axes: {}".format([type(q) for q in axis]))
            if any(d > len(x.shape) for d in axis):
                raise ValueError("axes (axis) must be < {}, currently are {}".format(len(x.shape), axis))
            if any(d < 0 for d in axis):
                axis = [j % len(x.shape) for j in axis]

            output_shape = [output_shape[it] for it in range(len(output_shape)) if it not in axis]
            # multiple dimensions
            if x.split is None:
                return dndarray.DNDarray(torch.mean(x._DNDarray__array, dim=axis),
                                         tuple(output_shape), x.dtype, x.split, x.device, x.comm)

            if x.split in axis:
                # merge in the direction of the split
                return reduce_means_elementwise(output_shape)
            else:
                # multiple dimensions which does *not* include the split axis
                # combine along the split axis
                return dndarray.DNDarray(torch.mean(x._DNDarray__array, dim=axis),
                                         tuple(output_shape), x.dtype,
                                         x.split if x.split < len(output_shape) else len(output_shape) - 1,
                                         x.device, x.comm)
        elif isinstance(axis, int):
            if axis >= len(x.shape):
                raise ValueError("axis (axis) must be < {}, currently is {}".format(len(x.shape), axis))
            axis = axis if axis > 0 else axis % len(x.shape)

            # only one axis given
            output_shape = [output_shape[it] for it in range(len(output_shape)) if it != axis]
            output_shape = output_shape if output_shape else (1, )

            if x.split is None:
                # return operations.__local_op(torch.mean, x, out=None, **{'dim': axis})
                return dndarray.DNDarray(torch.mean(x._DNDarray__array, dim=axis),
                                         tuple(output_shape), x.dtype, x.split, x.device, x.comm)
            elif axis == x.split:
                return reduce_means_elementwise(output_shape)
            else:
                # singular axis given (axis) not equal to split direction (x.split)
                return dndarray.DNDarray(torch.mean(x._DNDarray__array, dim=axis), tuple(output_shape), x.dtype,
                                         x.split if x.split < len(output_shape) else len(output_shape) - 1, x.device, x.comm)
        else:
            raise TypeError(
                'axis (axis) must be an int or a list, ht.DNDarray, torch.Tensor, or tuple, but was {}'.format(type(axis)))


def merge_means(mu1, n1, mu2, n2):
    """
    Function to merge two means by pairwise update.
    **Note** all tensors/arrays must be either the same size or individual values (can be mixed, i.e. n can be a float)

    Parameters
    ----------
    mu1 : ht.DNDarray, torch.tensor, float, int
        Calculated mean
    n1 : ht.DNDarray, torch.tensor, float
        number of elements used to calculate mu1
    mu2 : ht.DNDarray, torch.tensor, float, int
        Calculated mean
    n2 : ht.DNDarray, torch.tensor, float
        number of elements used to calculate mu2

    Returns
    -------
    combined_set_count : int
        Number of elements in the combined set.

    References
    ----------
    [1] J. Bennett, R. Grout, P. Pebay, D. Roe, D. Thompson, Numerically stable, single-pass, parallel statistics
        algorithms, IEEE International Conference on Cluster Computing and Workshops, 2009, Oct 2009, New Orleans, LA,
        USA.
    """
    return mu1 + n2 * ((mu2 - mu1) / (n1 + n2)), n1 + n2


def merge_vars(var1, mu1, n1, var2, mu2, n2, bessel=True):
    """
    Function to merge two variances by pairwise update.
    **Note** this is a parallel of the merge_means function
    **Note pt2.** all tensors/arrays must be either the same size or individual values

    Parameters
    ----------
    var1 : ht.DNDarray, torch.tensor, float, int
        Variance.
    mu1 : ht.DNDarray, torch.tensor, float, int
        Calculated mean.
    n1 : ht.DNDarray, torch.tensor, float, int
        Number of elements used to calculate mu1.
    var2 : ht.DNDarray, torch.tensor, float, int
        Variance.
    mu2 : ht.DNDarray, torch.tensor, float, int
        Calculated mean.
    n2 : ht.DNDarray, torch.tensor, float, int
        Number of elements used to calculate mu2.
    bessel : bool
        Flag for the use of the bessel correction

    Returns
    -------
    combined_set_count : int
        Number of elements in the combined set.

    References
    ----------
    [1] J. Bennett, R. Grout, P. Pebay, D. Roe, D. Thompson, Numerically stable, single-pass, parallel statistics
        algorithms, IEEE International Conference on Cluster Computing and Workshops, 2009, Oct 2009, New Orleans, LA,
        USA.
    """
    n = n1 + n2
    delta = mu2 - mu1
    if bessel:
        return (var1 * (n1 - 1) + var2 * (n2 - 1) + (delta ** 2) * n1 * n2 / n) / (n - 1), mu1 + n2 * (delta / (n1 + n2)), n
    else:
        return (var1 * n1 + var2 * n2 + (delta ** 2) * n1 * n2 / n) / n, mu1 + n2 * (delta / (n1 + n2)), n


def min(x, axis=None, out=None, keepdim=None):
    # TODO: initial : scalar, optional Issue #101
    """
    Return the minimum along a given axis.

    Parameters
    ----------
    a : ht.DNDarray
        Input data.
    axis : None or int or tuple of ints
        Axis or axes along which to operate. By default, flattened input is used.
        If this is a tuple of ints, the minimum is selected over multiple axes,
        instead of a single axis or all the axes as before.
    out : ht.DNDarray, optional
        Tuple of two output tensors (min, min_indices). Must be of the same shape and buffer length as the expected
        output. The maximum value of an output element. Must be present to allow computation on empty slice.

    Returns
    -------
    minimums : ht.DNDarray
        The minimums along a given axis.

    Examples
    --------
    >>> a = ht.float32([
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
            [10, 11, 12]
        ])
    >>> ht.min(a)
    tensor([1.])
    >>> ht.min(a, axis=0)
    tensor([[1., 2., 3.]])
    >>> ht.min(a, axis=1)
    tensor([[ 1.],
        [ 4.],
        [ 7.],
        [10.]])
    """
    def local_min(*args, **kwargs):
        result = torch.min(*args, **kwargs)
        if isinstance(result, tuple):
            return result[0]
        return result

    return operations.__reduce_op(x, local_min, MPI.MIN, axis=axis, out=out, keepdim=keepdim)


def mpi_argmax(a, b, _):
    lhs = torch.from_numpy(np.frombuffer(a, dtype=np.float64))
    rhs = torch.from_numpy(np.frombuffer(b, dtype=np.float64))

    # extract the values and minimal indices from the buffers (first half are values, second are indices)
    values = torch.stack((lhs.chunk(2)[0], rhs.chunk(2)[0],), dim=1)
    indices = torch.stack((lhs.chunk(2)[1], rhs.chunk(2)[1],), dim=1)

    # determine the minimum value and select the indices accordingly
    max, max_indices = torch.max(values, dim=1)
    result = torch.cat((max, indices[torch.arange(values.shape[0]), max_indices],))

    rhs.copy_(result)


MPI_ARGMAX = MPI.Op.Create(mpi_argmax, commute=True)


def mpi_argmin(a, b, _):
    lhs = torch.from_numpy(np.frombuffer(a, dtype=np.float64))
    rhs = torch.from_numpy(np.frombuffer(b, dtype=np.float64))

    # extract the values and minimal indices from the buffers (first half are values, second are indices)
    values = torch.stack((lhs.chunk(2)[0], rhs.chunk(2)[0],), dim=1)
    indices = torch.stack((lhs.chunk(2)[1], rhs.chunk(2)[1],), dim=1)

    # determine the minimum value and select the indices accordingly
    min, min_indices = torch.min(values, dim=1)
    result = torch.cat(
        (min, indices[torch.arange(values.shape[0]), min_indices],))

    rhs.copy_(result)


MPI_ARGMIN = MPI.Op.Create(mpi_argmin, commute=True)


def std(x, axis=None, bessel=True):
    """
    Calculates and returns the standard deviation of a tensor with the bessel correction.
    If a axis is given, the variance will be taken in that direction.

    Parameters
    ----------
    x : ht.DNDarray
        Values for which the std is calculated for.
    axis : None, int, defaults to None
        Axis which the mean is taken in. Default None calculates the standard deviation of all data items.
    bessel : bool, defaults to True
        Use the bessel correction when calculating the variance/std. Toggles between unbiased and biased calculation of
        the standard deviation.

    Returns
    -------
    stds : ht.DNDarray
        The std/s, if split, then split in the same direction as x.

    Examples
    --------
    >>> a = ht.random.randn(1,3)
    >>> a
    tensor([[ 0.3421,  0.5736, -2.2377]])
    >>> ht.std(a)
    tensor(1.5606)
    >>> a = ht.random.randn(4,4)
    >>> a
    tensor([[-1.0206,  0.3229,  1.1800,  1.5471],
            [ 0.2732, -0.0965, -0.1087, -1.3805],
            [ 0.2647,  0.5998, -0.1635, -0.0848],
            [ 0.0343,  0.1618, -0.8064, -0.1031]])
    >>> ht.std(a, 0)
    tensor([0.6157, 0.2918, 0.8324, 1.1996])
    >>> ht.std(a, 1)
    tensor([1.1405, 0.7236, 0.3506, 0.4324])
    >>> ht.std(a, 1, bessel=False)
    tensor([0.9877, 0.6267, 0.3037, 0.3745])
    """
    if not axis:
        return np.sqrt(var(x, axis, bessel))
    else:
        return exponential.sqrt(var(x, axis, bessel), out=None)


def var(x, axis=None, bessel=True):
    """
    Calculates and returns the variance of a tensor. If an axis is given, the variance will be taken in that direction.

    Parameters
    ----------
    x : ht.DNDarray
        Values for which the variance is calculated for.
    axis : None, int, defaults to None
        Axis which the variance is taken in. Default None calculates the variance of all data items.
    bessel : bool, defaults to True
        Use the bessel correction when calculating the variance/std. Toggles between unbiased and biased calculation of
        the variance.

    Returns
    -------
    variances : ht.DNDarray
        The var/s, if split, then split in the same direction as x.

    Examples
    --------
    >>> a = ht.random.randn(1,3)
    >>> a
    tensor([[-1.9755,  0.3522,  0.4751]])
    >>> ht.var(a)
    tensor(1.9065)

    >>> a = ht.random.randn(4,4)
    >>> a
    tensor([[-0.8665, -2.6848, -0.0215, -1.7363],
            [ 0.5886,  0.5712,  0.4582,  0.5323],
            [ 1.9754,  1.2958,  0.5957,  0.0418],
            [ 0.8196, -1.2911, -0.2026,  0.6212]])
    >>> ht.var(a, 1)
    tensor([1.3092, 0.0034, 0.7061, 0.9217])
    >>> ht.var(a, 0)
    tensor([1.3624, 3.2563, 0.1447, 1.2042])
    >>> ht.var(a, 0, bessel=True)
    tensor([1.3624, 3.2563, 0.1447, 1.2042])
    >>> ht.var(a, 0, bessel=False)
    tensor([1.0218, 2.4422, 0.1085, 0.9032])
    """
    if not isinstance(bessel, bool):
        raise TypeError('bessel must be a boolean, currently is {}'.format(type(bessel)))

    def reduce_vars_elementwise(output_shape_i):
        """
        Function to combine the calculated vars together. This does an element-wise update of the calculated vars to
        merge them together using the merge_vars function. This function operates using x from the var function
        parameters.

        Parameters
        ----------
        output_shape_i : iterable
            Iterable with the dimensions of the output of the var function.

        Returns
        -------
        variances : ht.DNDarray
            The calculated variances.
        """

        if x.lshape[x.split] != 0:
            mu = operations.__local_op(torch.mean, x, out=None, dim=axis)
            var = torch.var(x._DNDarray__array, out=None, dim=axis, unbiased=bessel)
        else:
            mu = factories.zeros(output_shape_i)
            var = factories.zeros(output_shape_i)

        n_for_merge = factories.zeros(x.comm.size)
        n2 = factories.zeros(x.comm.size)
        n2[x.comm.rank] = x.lshape[x.split]
        x.comm.Allreduce(n2, n_for_merge, MPI.SUM)

        var_shape = list(var.shape) if list(var.shape) else [1]

        var_tot = factories.zeros(([x.comm.size, 2] + var_shape))
        n_tot = factories.zeros(x.comm.size)
        var_tot[x.comm.rank, 0, :] = var
        var_tot[x.comm.rank, 1, :] = mu
        n_tot[x.comm.rank] = float(x.lshape[x.split])
        x.comm.Allreduce(MPI.IN_PLACE, var_tot, MPI.SUM)
        x.comm.Allreduce(MPI.IN_PLACE, n_tot, MPI.SUM)

        for i in range(1, x.comm.size):
            var_tot[0, 0, :], var_tot[0, 1, :], n_tot[0] = merge_vars(var_tot[0, 0, :], var_tot[0, 1, :], n_tot[0],
                                                                      var_tot[i, 0, :], var_tot[i, 1, :], n_tot[i])
        return var_tot[0, 0, :]
    # ----------------------------------------------------------------------------------------------------
    if axis is None:  # no axis given
        if not x.is_distributed():  # not distributed (full tensor on one node)
            ret = torch.var(x._DNDarray__array, unbiased=bessel)
            return dndarray.DNDarray(ret, tuple(ret.shape), x.dtype, None, x.device, x.comm)

        else:  # case for full matrix calculation (axis is None)
            if x.lshape[x.split] != 0:
                mu_in = operations.__local_op(torch.mean, x, out=None)
                var_in = operations.__local_op(torch.var, x, out=None, unbiased=bessel)
                if torch.isnan(var_in._DNDarray__array):
                    var_in = 0.0
            else:
                mu_in = 0
                var_in = 0
            n = x.lnumel
            var_tot = factories.zeros((x.comm.size, 3))
            var_proc = factories.zeros((x.comm.size, 3))
            var_proc[x.comm.rank][0] = var_in
            var_proc[x.comm.rank][1] = mu_in
            var_proc[x.comm.rank][2] = float(n)
            x.comm.Allreduce(var_proc, var_tot, MPI.SUM)

            for i in range(1, x.comm.size):
                merged = merge_vars(var_tot[0, 0], var_tot[0, 1], var_tot[0, 2], var_tot[i, 0], var_tot[i, 1], var_tot[i, 2])
                var_tot[0, 0] = merged[0]
                var_tot[0, 1] = merged[1]
                var_tot[0, 2] = merged[2]

            return var_tot[0][0]

    else:  # axis is given
        # case for var in one dimension
        output_shape = list(x.shape)
        if isinstance(axis, int):
            if axis >= len(x.shape):
                raise ValueError("axis must be < {}, currently is {}".format(len(x.shape), axis))
            axis = axis if axis > 0 else axis % len(x.shape)
            # only one axis given
            output_shape = [output_shape[it] for it in range(len(output_shape)) if it != axis]
            output_shape = output_shape if output_shape else (1,)

            if x.split is None:  # x is *not* distributed -> no need to distributed
                return dndarray.DNDarray(torch.var(x._DNDarray__array, dim=axis, unbiased=bessel),
                                         tuple(output_shape), x.dtype, None, x.device, x.comm)
            elif axis == x.split:  # x is distributed and axis chosen is == to split
                return reduce_vars_elementwise(output_shape)
            else:
                # singular axis given (axis) not equal to split direction (x.split)
                lcl = torch.var(x._DNDarray__array, dim=axis, keepdim=True)
                return dndarray.DNDarray(lcl, tuple(output_shape), x.dtype, x.split if x.split < len(output_shape) else len(output_shape) - 1,
                                         x.device, x.comm)
        else:
            raise TypeError(
                'axis (axis) must be an int, currently is {}. Check if multidim var is available in PyTorch'.format(type(axis)))



import itertools
import numpy as np


def broadcast_shape(shape_a, shape_b):
    """
    Infers, if possible, the broadcast output shape of two operands a and b. Inspired by stackoverflow post:
    https://stackoverflow.com/questions/24743753/test-if-an-array-is-broadcastable-to-a-shape

    Parameters
    ----------
    shape_a : tuple of ints
        shape of operand a
    shape_b : tuple of ints
        shape of operand b

    Returns
    -------
    broadcast_shape : tuple of ints
        the broadcast shape

    Raises
    -------
    ValueError
        If the two shapes cannot be broadcast.

    Examples
    -------
    >>> broadcast_shape((5,4),(4,))
    (5,4)

    >>> broadcast_shape((1,100,1),(10,1,5))
    (10,100,5)

    >>> broadcast_shape((8,1,6,1),(7,1,5,))
    (8,7,6,5))

    >>> broadcast_shape((2,1),(8,4,3))
    ValueError
    """

    it = itertools.zip_longest(shape_a[::-1], shape_b[::-1], fillvalue=1)
    resulting_shape = max(len(shape_a), len(shape_b)) * [None]
    for i, (a, b) in enumerate(it):
        if a == 1 or b == 1 or a == b:
            resulting_shape[i] = max(a, b)
        else:
            raise ValueError('operands could not be broadcast, input shapes {} {}'.format(shape_a, shape_b))

    return tuple(resulting_shape[::-1])


def sanitize_axis(shape, axis):
    """
    Checks conformity of an axis with respect to a given shape. The axis will be converted to its positive equivalent
    and is checked to be within bounds

    Parameters
    ----------
    shape : tuple of ints
        shape of an array
    axis : ints or tuple of ints
        the axis to be sanitized

    Returns
    -------
    sane_axis : int or tuple of ints
        the sane axis

    Raises
    -------
    ValueError
        if the axis cannot be sanitized, i.e. out of bounds.
    TypeError
        if the the axis is not integral.

    Examples
    -------
    >>> sanitize_axis((5,4,4),1)
    1

    >>> sanitize_axis((5,4,4),-1)
    2

    >>> sanitize_axis((5, 4), (1,))
    (1,)

    >>> sanitize_axis((5, 4), 1.0)
    TypeError
    """
    if axis is not None:
        if not isinstance(axis, int) and not isinstance(axis, tuple):
            raise TypeError('axis must be None or int or tuple, but was {}'.format(type(axis)))
    if isinstance(axis, tuple):
        axis = tuple(dim+len(shape) if dim < 0 else dim for dim in axis)
        for dim in axis:
            if dim < 0 or dim >= len(shape):
                raise ValueError('axis {} is out of bounds for shape {}'.format(axis, shape))
        return axis

    if axis is None or 0 <= axis < len(shape):
        return axis
    elif axis < 0:
        axis += len(shape)

    if axis < 0 or axis >= len(shape):
        raise ValueError('axis {} is out of bounds for shape {}'.format(axis, shape))

    return axis


def sanitize_shape(shape):
    """
    Verifies and normalizes the given shape.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of an array.

    Returns
    -------
    sane_shape : tuple of ints
        The sanitized shape.

    Raises
    -------
    ValueError
        If the shape contains illegal values, e.g. negative numbers.
    TypeError
        If the given shape is neither and int or a sequence of ints.

    Examples
    --------
    >>> sanitize_shape(3)
    (3,)

    >>> sanitize_shape([1, 2, 3])
    (1, 2, 3,)

    >>> sanitize_shape(1.0)
    TypeError
    """
    shape = (shape,) if not hasattr(shape, '__iter__') else tuple(shape)

    for dimension in shape:
        if issubclass(type(dimension), np.integer):
            dimension = int(dimension)
        if not isinstance(dimension, int):
            raise TypeError('expected sequence object with length >= 0 or a single integer')
        if dimension < 0:
            raise ValueError('negative dimensions are not allowed')

    return shape



import torch

from .communication import MPI_WORLD


__all__ = [
    'cpu',
    'get_backend',
    'sanitize_device',
    'use_backend'
]


class Device:
    """
    Implements a compute device.

    HeAT can run computations on different compute devices or backends. A device describes the device type and id on
    which said computation should be carried out.

    Parameters
    ----------
    device_type : str
        represents HeAT's device name
    device_id : int
        the device id
    torch_device : str
        the corresponding PyTorch device type

    Examples
    --------
    >>> # array on cpu
    >>> cpu_array = ht.ones((2, 3), device=ht.cpu)
    >>> # array on gpu
    >>> gpu_array = ht.ones((2, 3), device=ht.gpu)
    device(cpu:0)
    """
    def __init__(self, device_type, device_id, torch_device):
        self.__device_type = device_type
        self.__device_id = device_id
        self.__torch_device = torch_device

    @property
    def device_type(self):
        return self.__device_type

    @property
    def device_id(self):
        return self.__device_id

    @property
    def torch_device(self):
        return self.__torch_device

    def __repr__(self):
        return 'device({})'.format(self.__str__())

    def __str__(self):
        return '{}:{}'.format(self.device_type, self.device_id)


# create a CPU device singleton
cpu = Device('cpu', 0, 'cpu:0')

# define the default device to be the CPU
__default_device = cpu
# add a device string for the CPU device
__device_mapping = {
    cpu.device_type: cpu
}

# add gpu support if available
if torch.cuda.device_count() > 0:
    # GPUs are assigned round-robin to the MPI processes
    gpu_id = MPI_WORLD.rank % torch.cuda.device_count()
    # create a new GPU device
    gpu = Device('gpu', gpu_id, 'cuda:{}'.format(gpu_id))
    # add a GPU device string
    __device_mapping[gpu.device_type] = gpu
    # the GPU device should be exported as global symbol
    __all__.append('gpu')


def get_backend():
    """
    Retrieves the currently globally set default device.

    Returns
    -------
    defaults device : Device
        The currently set default device.
    """
    return __default_device


def sanitize_device(device):
    """
    Sanitizes a device or device identifier, i.e. checks whether it is already an instance of Device or a string with
    known device identifier and maps it to a proper Device.

    Parameters
    ----------
    device : str, Device or None
        The device to be sanitized

    Returns
    -------
    sanitized_device : Device
        The matching Device instance

    Raises
    ------
    ValueError
        If the given device id is not recognized
    """
    if device is None:
        return get_backend()

    if isinstance(device, Device):
        return device

    try:
        return __device_mapping[device.strip().lower()]
    except (AttributeError, KeyError, TypeError):
        raise ValueError('Unknown device, must be one of {}'.format(', '.join(__device_mapping.keys())))


def use_backend(device=None):
    """
    Sets the globally used default device.

    Parameters
    ----------
    device : str, Device or None
        The device to be set
    """
    global __default_device
    __default_device = sanitize_device(device)



from distutils.core import setup
import sys
sys.path.append('./heat/core')
import version

print(version, dir(version))

with open('README.md', 'r') as handle:
    long_description = handle.read()

# with open('./heat/core/version.py') as handle:
#     exec(handle.read())
#     print(dir())

setup(
    name='heat',
    packages=['heat', 'heat.core', 'heat.ml', 'heat.ml.cluster'],
    data_files=['README.md', 'LICENSE'],
    version=version.__version__,
    description='A framework for high performance data analytics and machine learning.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Helmholtz Association',
    author_email='martin.siggel@dlr.de',
    url='https://github.com/helmholtz-analytics/heat',
    keywords=['data', 'analytics', 'tensors', 'distributed', 'gpu'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering'
    ],
    install_requires=[
        'mpi4py>=3.0.0',
        'numpy>=1.13.0',
        'torch>=1.0.0'
    ],
    extras_require={
        'hdf5':  ['h5py>=2.8.0'],
        'netcdf': ['netCDF4>=1.4.0'],
    }
)



import os.path
import torch
import warnings
from .stride_tricks import sanitize_axis
from .communication import MPI, MPI_WORLD
from . import devices
from . import types

__VALID_WRITE_MODES = frozenset(['w', 'a', 'r+'])
__HDF5_EXTENSIONS = frozenset(['.h5', '.hdf5'])
__NETCDF_EXTENSIONS = frozenset(['.nc', '.nc4', 'netcdf'])
__NETCDF_DIM_TEMPLATE = '{}_dim_{}'

__all__ = [
    'load',
    'save'
]

try:
    import h5py
except ImportError:
    # HDF5 support is optional
    def supports_hdf5():
        return False
else:
    # warn the user about serial hdf5
    if not h5py.get_config().mpi and MPI_WORLD.rank == 0:
        warnings.warn('h5py does not support parallel I/O, falling back to slower serial I/O', ImportWarning)

    # add functions to exports
    __all__.extend([
        'load_hdf5',
        'save_hdf5'
    ])

    def supports_hdf5():
        return True


    def load_hdf5(path, dataset, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
        """
        Loads data from an HDF5 file. The data may be distributed among multiple processing nodes via the split flag.

        Parameters
        ----------
        path : str
            Path to the HDF5 file to be read.
        dataset : str
            Name of the dataset to be read.
        dtype : ht.dtype
            Data type of the resulting array; default: ht.float32.
        split : int, optional
            The axis along which the data is distributed among the processing cores.
        device : None or str, optional
            The device id on which to place the data, defaults to globally set default device.
        comm : ht.Communication, optional
            The communication to use for the data distribution. defaults to MPI_COMM_WORLD.

        Returns
        -------
        out : ht.DNDarray
            Data read from the HDF5 file.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type

        Examples
        --------
        >>> a = ht.load_hdf5('data.h5', dataset='DATA')
        >>> a.shape
        (5,)
        >>> a.lshape
        (5,)
        >>> b = ht.load_hdf5('data.h5', dataset='DATA', split=0)
        >>> b.shape
        (5,)
        >>> b.lshape
        (3,)
        """
        if not isinstance(path, str):
            raise TypeError('path must be str, not {}'.format(type(path)))
        if not isinstance(dataset, str):
            raise TypeError('dataset must be str, not {}'.format(type(dataset)))
        if split is not None and not isinstance(split, int):
            raise TypeError('split must be None or int, not {}'.format(type(split)))

        # infer the type and communicator for the loaded array
        dtype = types.canonical_heat_type(dtype)
        # determine the device the data will be placed on
        device = devices.sanitize_device(device)

        # actually load the data from the HDF5 file
        with h5py.File(path, 'r') as handle:
            data = handle[dataset]
            gshape = tuple(data.shape)
            dims = len(gshape)
            split = sanitize_axis(gshape, split)
            _, _, indices = comm.chunk(gshape, split)
            if split is None: 
                data = torch.tensor(data[indices], dtype=dtype.torch_type(), device=device.torch_device)
            elif indices[split].stop > indices[split].start: 
                data = torch.tensor(data[indices], dtype=dtype.torch_type(), device=device.torch_device)
            else: 
                warnings.warn('More MPI ranks are used then the length of splitting dimension!')
                slice1 = tuple(slice(0, gshape[i]) if i != split else slice(0, 1) for i in range(dims)) 
                slice2 = tuple(slice(0, gshape[i]) if i != split else slice(0, 0) for i in range(dims)) 
                data = torch.tensor(data[slice1], dtype=dtype.torch_type(), device=device.torch_device)
                data = data[slice2]
                
            return dndarray.DNDarray(data, gshape, dtype, split, device, comm)


    def save_hdf5(data, path, dataset, mode='w', **kwargs):
        """
        Saves data to an HDF5 file. Attempts to utilize parallel I/O if possible.

        Parameters
        ----------
        data : ht.DNDarray
            The data to be saved on disk.
        path : str
            Path to the HDF5 file to be written.
        dataset : str
            Name of the dataset the data is saved to.
        mode : str, one of 'w', 'a', 'r+'
            File access mode
        kwargs : dict
            additional arguments passed to the created dataset.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type.
        ValueError
            If the access mode is not understood.

        Examples
        --------
        >>> a_range = ht.arange(100, split=0)
        >>> ht.save_hdf5(a_range, 'data.h5', dataset='DATA')
        """
        if not isinstance(data, dndarray.DNDarray):
            raise TypeError('data must be heat tensor, not {}'.format(type(data)))
        if not isinstance(path, str):
            raise TypeError('path must be str, not {}'.format(type(path)))
        if not isinstance(dataset, str):
            raise TypeError('dataset must be str, not {}'.format(type(path)))

        # we only support a subset of possible modes
        if mode not in __VALID_WRITE_MODES:
            raise ValueError('mode was {}, not in possible modes {}'.format(mode, __VALID_WRITE_MODES))

        # chunk the data, if no split is set maximize parallel I/O and chunk first axis
        is_split = data.split is not None
        _, _, slices = data.comm.chunk(data.gshape, data.split if is_split else 0)

        # attempt to perform parallel I/O if possible
        if h5py.get_config().mpi:
            with h5py.File(path, mode, driver='mpio', comm=data.comm.handle) as handle:
                dset = handle.create_dataset(dataset, data.shape, **kwargs)
                dset[slices] = data._DNDarray__array.cpu() if is_split else data._DNDarray__array[slices].cpu()

        # otherwise a single rank only write is performed in case of local data (i.e. no split)
        elif data.comm.rank == 0:
            with h5py.File(path, mode) as handle:
                dset = handle.create_dataset(dataset, data.shape, **kwargs)
                if is_split:
                    dset[slices] = data._DNDarray__array.cpu()
                else:
                    dset[...] = data._DNDarray__array.cpu()

            # ping next rank if it exists
            if is_split and data.comm.size > 1:
                data.comm.Isend([None, 0, MPI.INT], dest=1)
                data.comm.Recv([None, 0, MPI.INT], source=data.comm.size - 1)

        # no MPI, but split data is more tricky, we have to serialize the writes
        elif is_split:
            # wait for the previous rank to finish writing its chunk, then write own part
            data.comm.Recv([None, 0, MPI.INT], source=data.comm.rank - 1)
            with h5py.File(path, 'r+') as handle:
                handle[dataset][slices] = data._DNDarray__array.cpu()

            # ping the next node in the communicator, wrap around to 0 to complete barrier behavior
            next_rank = (data.comm.rank + 1) % data.comm.size
            data.comm.Isend([None, 0, MPI.INT], dest=next_rank)

try:
    import netCDF4 as nc
except ImportError:
    def supports_netcdf():
        return False
else:
    __nc_has_par = nc.__dict__.get('__has_parallel4_support__', False) or \
                   nc.__dict__.get('__has_pnetcdf_support__', False) or \
                   nc.__dict__.get('__has_nc_par__', False)

    # warn the user about serial netcdf
    if not __nc_has_par and MPI_WORLD.rank == 0:
        warnings.warn('netCDF4 does not support parallel I/O, falling back to slower serial I/O', ImportWarning)

    # add functions to visible exports
    __all__.extend([
        'load_netcdf',
        'save_netcdf'
    ])

    def supports_netcdf():
        return True

    def load_netcdf(path, variable, dtype=types.float32, split=None, device=None, comm=MPI_WORLD):
        """
        Loads data from a NetCDF4 file. The data may be distributed among multiple processing nodes via the split flag.

        Parameters
        ----------
        path : str
            Path to the NetCDF4 file to be read.
        variable : str
            Name of the variable to be read.
        dtype : ht.dtype
            Data type of the resulting array; default: ht.float32.
        split : int, optional
            The axis along which the data is distributed among the processing cores.
        comm : ht.Communication, optional
            The communication to use for the data distribution. defaults to MPI_COMM_WORLD.
        device : None or str, optional
            The device id on which to place the data, defaults to globally set default device.

        Returns
        -------
        out : ht.DNDarray
            Data read from the NetCDF4 file.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type

        Examples
        --------
        >>> a = ht.load_netcdf('data.nc', variable='DATA')
        >>> a.shape
        (5,)
        >>> a.lshape
        (5,)
        >>> b = ht.load_netcdf('data.nc', variable='DATA', split=0)
        >>> b.shape
        (5,)
        >>> b.lshape
        (3,)
        """
        if not isinstance(path, str):
            raise TypeError('path must be str, not {}'.format(type(path)))
        if not isinstance(variable, str):
            raise TypeError('dataset must be str, not {}'.format(type(variable)))
        if split is not None and not isinstance(split, int):
            raise TypeError('split must be None or int, not {}'.format(type(split)))

        # infer the canonical heat datatype
        dtype = types.canonical_heat_type(dtype)
        # determine the device the data will be placed on
        device = devices.sanitize_device(device)

        # actually load the data
        with nc.Dataset(path, 'r', parallel=__nc_has_par, comm=comm.handle) as handle:
            data = handle[variable][:]

            # prepare meta information
            gshape = tuple(data.shape)
            split = sanitize_axis(gshape, split)

            # chunk up the data portion
            _, local_shape, indices = comm.chunk(gshape, split)
            if split is None or local_shape[split] > 0:
                data = torch.tensor(data[indices], dtype=dtype.torch_type(), device=device.torch_device)
            else:
                data = torch.empty(local_shape, dtype=dtype.torch_type(), device=device.torch_device)

            return dndarray.DNDarray(data, gshape, dtype, split, device, comm)


    def save_netcdf(data, path, variable, mode='w', **kwargs):
        """
        Saves data to a netCDF4 file. Attempts to utilize parallel I/O if possible.

        Parameters
        ----------
        data : ht.DNDarray
            The data to be saved on disk.
        path : str
            Path to the netCDF4 file to be written.
        variable : str
            Name of the variable the data is saved to.
        mode : str, one of 'w', 'a', 'r+'
            File access mode
        kwargs : dict
            additional arguments passed to the created dataset.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type.
        ValueError
            If the access mode is not understood.

        Examples
        --------
        >>> a_range = ht.arange(100, split=0)
        >>> ht.save_netcdf(a_range, 'data.nc', dataset='DATA')
        """
        if not isinstance(data, dndarray.DNDarray):
            raise TypeError('data must be heat tensor, not {}'.format(type(data)))
        if not isinstance(path, str):
            raise TypeError('path must be str, not {}'.format(type(path)))
        if not isinstance(variable, str):
            raise TypeError('variable must be str, not {}'.format(type(path)))

        # we only support a subset of possible modes
        if mode not in __VALID_WRITE_MODES:
            raise ValueError('mode was {}, not in possible modes {}'.format(mode, __VALID_WRITE_MODES))

        # chunk the data, if no split is set maximize parallel I/O and chunk first axis
        is_split = data.split is not None
        _, _, slices = data.comm.chunk(data.gshape, data.split if is_split else 0)

        # attempt to perform parallel I/O if possible
        if __nc_has_par:
            with nc.Dataset(path, mode, parallel=True, comm=data.comm.handle) as handle:
                dimension_names = []
                for dimension, elements in enumerate(data.shape):
                    name = __NETCDF_DIM_TEMPLATE.format(variable, dimension)
                    handle.createDimension(name, elements)
                    dimension_names.append(name)

                var = handle.createVariable(variable, data.dtype.char(), dimension_names, **kwargs)
                var[slices] = data._DNDarray__array.cpu() if is_split else data._DNDarray__array[slices].cpu()

        # otherwise a single rank only write is performed in case of local data (i.e. no split)
        elif data.comm.rank == 0:
            with nc.Dataset(path, mode) as handle:
                dimension_names = []
                for dimension, elements in enumerate(data.shape):
                    name = __NETCDF_DIM_TEMPLATE.format(variable, dimension)
                    handle.createDimension(name, elements)
                    dimension_names.append(name)

                var = handle.createVariable(variable, data.dtype.char(), tuple(dimension_names), **kwargs)
                if is_split:
                    var[slices] = data._DNDarray__array.cpu()
                else:
                    var[:] = data._DNDarray__array.cpu()

            # ping next rank if it exists
            if is_split and data.comm.size > 1:
                data.comm.Isend([None, 0, MPI.INT], dest=1)
                data.comm.Recv([None, 0, MPI.INT], source=data.comm.size - 1)

        # no MPI, but data is split, we have to serialize the writes
        elif is_split:
            # wait for the previous rank to finish writing its chunk, then write own part
            data.comm.Recv([None, 0, MPI.INT], source=data.comm.rank - 1)
            with nc.Dataset(path, 'r+') as handle:
                handle[variable][slices] = data._DNDarray__array.cpu()

            # ping the next node in the communicator, wrap around to 0 to complete barrier behavior
            next_rank = (data.comm.rank + 1) % data.comm.size
            data.comm.Isend([None, 0, MPI.INT], dest=next_rank)


def load(path, *args, **kwargs):
    """
    Attempts to load data from a file stored on disk. Attempts to auto-detect the file format by determining the
    extension.

    Parameters
    ----------
    path : str
        Path to the file to be read.
    args/kwargs : list/dict
        additional options passed to the particular functions.

    Returns
    -------
    out : ht.DNDarray
        Data read from the file.

    Raises
    -------
    ValueError
        If the file extension is not understood or known.

    Examples
    --------
    >>> ht.load('data.h5', dataset='DATA')
    tensor([ 1.0000,  2.7183,  7.3891, 20.0855, 54.5981])
    >>> ht.load('data.nc', variable='DATA')
    tensor([ 1.0000,  2.7183,  7.3891, 20.0855, 54.5981])
    """
    if not isinstance(path, str):
        raise TypeError('Expected path to be str, but was {}'.format(type(path)))
    extension = os.path.splitext(path)[-1].strip().lower()

    if supports_hdf5() and extension in __HDF5_EXTENSIONS:
        return load_hdf5(path, *args, **kwargs)
    elif supports_netcdf() and extension in __NETCDF_EXTENSIONS:
        return load_netcdf(path, *args, **kwargs)
    else:
        raise ValueError('Unsupported file extension {}'.format(extension))


def save(data, path, *args, **kwargs):
    """
    Attempts to save data from a tensor to disk. Attempts to auto-detect the file format by determining the extension.

    Parameters
    ----------
    data : ht.DNDarray
        The tensor holding the data to be stored
    path : str
        Path to the file to be stored.
    args/kwargs : list/dict
        additional options passed to the particular functions.

    Raises
    -------
    ValueError
        If the file extension is not understood or known.

    Examples
    --------
    >>> a_range = ht.arange(100, split=0)
    >>> ht.save(a_range, 'data.h5', 'DATA', mode='a')
    >>> ht.save(a_range, 'data.nc', 'DATA', mode='w')
    """
    if not isinstance(path, str):
        raise TypeError('Expected path to be str, but was {}'.format(type(path)))
    extension = os.path.splitext(path)[-1].strip().lower()

    if supports_hdf5() and extension in __HDF5_EXTENSIONS:
        save_hdf5(data, path, *args, **kwargs)
    elif supports_netcdf() and extension in __NETCDF_EXTENSIONS:
        save_netcdf(data, path, *args, **kwargs)
    else:
        raise ValueError('Unsupported file extension {}'.format(extension))


# tensor is imported at the very end to break circular dependency
from . import dndarray



from . import dndarray

__all__ = [
    'copy'
]


def copy(a):
    """
    Return an array copy of the given object.

    Parameters
    ----------
    a : ht.DNDarray
        Input data to be copied.

    Returns
    -------
    copied : ht.DNDarray
        A copy of the original
    """
    if not isinstance(a, dndarray.DNDarray):
        raise TypeError('input needs to be a tensor')
    return dndarray.DNDarray(a._DNDarray__array.clone(), a.shape, a.dtype, a.split, a.device, a.comm)
